/* Copyright Sun Microsystems, 2006, all rights reserved 

   Written by Ramon Garcia Fernandez with the finantial support of 
   Google Summer of code program 2006 */


#include <com/sun/star/uno/Reference.hxx>
#include <com/sun/star/uno/XComponentContext.hpp>
#include <com/sun/star/lang/XMultiComponentFactory.hpp>
#include <com/sun/star/uno/XInterface.hpp>

#include <com/sun/star/uno/DeploymentException.hpp>


#include <rtl/ustring.hxx>


namespace create_instance {
using com::sun::star::uno::Reference;
using com::sun::star::uno::XComponentContext;
using com::sun::star::lang::XMultiComponentFactory;
using com::sun::star::uno::XInterface;
using rtl::OUString;

using com::sun::star::uno::DeploymentException;



/* A generic function to create an instance of a service.
   Please ensure that the template argument T is unique 
   for the service that you are using, so that only one template instance
   is created for each service, and this instance is not shared
   with other services 

   This code is based in the implementation of obtaining introspection in Basic code */

template<typename T>
Reference<T> createInstance(Reference<XComponentContext> xContext, const char* name)
{
	static Reference<T> instance;

	if( !instance.is() )
	{
		// Service holen
		Reference<XMultiComponentFactory> xFactory = xContext->getServiceManager();
		if ( xFactory.is() )
		{
			Reference<XInterface> xI = xFactory->createInstanceWithContext( rtl::OUString::createFromAscii(name), xContext);
			if (xI.is())
				instance = Reference<T>::query(xI);
				//xI->queryInterface( ::getCppuType( (const Reference< XIntrospection > *)0 ), xIntrospection );
		}
	}
	OSL_ENSURE(instance.is(), "### Instance not accessable!?" );
	if( !instance.is() )
	{
		throw DeploymentException(
			OUString( RTL_CONSTASCII_USTRINGPARAM("instance creation failed") ),
			Reference<XInterface>() );
	}
	return instance;
}

}

using create_instance::createInstance;
