#include "objectbrowser_impl.hxx"

#include "browser_lang.hxx"
#include "create_instance.hxx"
#include "url.hxx"
#include "field_data.hxx"

#include <com/sun/star/beans/MethodConcept.hpp>
#include <com/sun/star/beans/PropertyAttribute.hpp>
#include <com/sun/star/lang/XMultiComponentFactory.hpp>

#include <utility>
#include <rtl/ustrbuf.hxx>

#include <vcl/unohelp2.hxx>

namespace objectbrowser_impl {

using namespace ::com::sun::star::reflection;
using namespace ::com::sun::star::uno;
using namespace ::com::sun::star::beans;
using namespace ::com::sun::star::lang;
using com::sun::star::datatransfer::clipboard::XClipboard;
    
using namespace ::rtl;

static Reference<XIdlReflection> getCoreReflectionImpl(XComponentContext& xContext);

static Reference<XIntrospection> getIntrospectionImpl(Reference<XComponentContext> xContext);

static Reference<XIdlClass> TypeToIdlClass(const Type& rType, XComponentContext& xContext);



void createMethodMap(struct ObjectBrowserImpl::object_data& od);

void insertInterface(SvTreeListBox& t, SvLBoxEntry* parent, const Reference<XIdlClass>& c, struct ObjectBrowserImpl::object_data& od, ObjectBrowserLang current_lang);

static void ExpandRec(SvTreeListBox& t, SvLBoxEntry* entry)
{
    if (t.GetModel()->HasChilds(entry)) {
	t.Expand(entry);
	SvTreeEntryList* children = t.GetModel()->GetChildList(entry);
	for (SvLBoxEntry* child = (SvLBoxEntry*) children->First(); child != NULL; 
	     child = (SvLBoxEntry*) children->Next()) {
	    ExpandRec(t, child);
	}
    }
    
}

PopupMenu* ObjectBrowserImpl::MyTreeListBox::CreateContextMenu()
{
    PopupMenu* m = new PopupMenu();
    SvLBoxEntry* e = FirstSelected();
    FieldData* field_data = static_cast<FieldData*>(e->GetFirstItem(FIELD_DATA_ITEM_ID));
    m->InsertItem(0, XubString(RTL_CONSTASCII_USTRINGPARAM("Copy")));
    if (field_data != NULL) {
	if (field_data->InFlatList()) {
	    m->InsertItem(1, XubString(RTL_CONSTASCII_USTRINGPARAM("View Interface")));
	}
	m->InsertItem(2, XubString(RTL_CONSTASCII_USTRINGPARAM("IDL Reference")));
    }
    return m;
}

void ObjectBrowserImpl::MyTreeListBox::ExcecuteContextMenuAction(USHORT selected_entry)
{
    SvLBoxEntry* e = FirstSelected();
    FieldData*  field_data;
    switch (selected_entry) {
	case 0: 
	{
	    Reference<XClipboard> aClipboard(GetClipboard());
	    vcl::unohelper::TextDataObject::CopyStringTo(GetEntryText(e), aClipboard );
	}
	break;
	case 1: 
	{
	    field_data = static_cast<FieldData*>(e->GetFirstItem(FIELD_DATA_ITEM_ID));
	    struct object_data& od = * static_cast<struct object_data*>(field_data->GetObjectData());
	    interface_to_entry_t::iterator ie = od.interface_to_entry.find(field_data->GetInterface());
	    if (ie != od.interface_to_entry.end()) {
		SvLBoxEntry* entry = ie->second;
		MakeVisible(entry);
		Select(entry, TRUE);
		ExpandRec(*this, entry);
	    }
	}
	break;
	case 2:
	    field_data = static_cast<FieldData*>(e->GetFirstItem(FIELD_DATA_ITEM_ID));
	    OpenURL(outer->m_xContext, field_data->GetURL());
	    break;
    }
}


ObjectBrowserImpl::MyWin::MyWin(ObjectBrowserImpl* the_outer):
    WorkWindow(NULL, WB_STDWORK),
    tree(this, WB_BORDER | WB_3DLOOK | WB_HASBUTTONS | WB_HASLINES | WB_HSCROLL | WB_TABSTOP
	 | WB_HASLINESATROOT | WB_HASBUTTONSATROOT, the_outer),
    lang_uno(this, 0), lang_basic(this, 0),
    outer(the_outer)
{
    tree.SetNodeDefaultImages( );
    tree.Show();
    lang_uno.SetText(XubString(RTL_CONSTASCII_USTRINGPARAM("UNO syntax")));
    lang_uno.Show();
    lang_uno.SetToggleHdl(LINK(this, ObjectBrowserImpl::MyWin, SelectLang));
    lang_basic.Show();
    lang_basic.SetText(XubString(RTL_CONSTASCII_USTRINGPARAM("Basic syntax")));
    Resize();
    SetText(XubString(RTL_CONSTASCII_USTRINGPARAM("Object browser")));
    lang_uno.SetState(TRUE);
    Show();
}
    
IMPL_LINK(ObjectBrowserImpl::MyWin, SelectLang, void*, EMPTYARG)
{
    ObjectBrowserLang lang;
    if (lang_uno.IsChecked()) {
	lang = ObjectBrowserLang_UNO;
    } else {
	lang = ObjectBrowserLang_BASIC;
    }
    outer->setlang(lang);
    return 0;
}

void ObjectBrowserImpl::setlang(ObjectBrowserLang lang) throw (RuntimeException)
{
    if (lang == current_lang) 
	return;
    Guard<Mutex> guard(mutex);
    current_lang = lang;
    std::hash_map<OUString, object_data, OUStringHash>::iterator i;
    for (i = objects.begin(); i != objects.end(); i++) {
	struct object_data& od = i->second;
	SvLBoxTreeList* model = tree.GetModel();
	while (model->HasChilds(od.tree_item)) {
	    SvLBoxEntry* child = model->FirstChild(od.tree_item);
	    model->Remove(child);
	}
	fillTree(od);
    }
    switch (lang) {
	case ObjectBrowserLang_UNO:
	    if (!w.lang_uno.IsChecked()) {
		w.lang_uno.Check(TRUE);
	    }
	    break;
	case ObjectBrowserLang_BASIC:
	    if (!w.lang_basic.IsChecked()) {
		w.lang_basic.Check(TRUE);
	    }
	    break;
    }
}

ObjectBrowserLang SAL_CALL ObjectBrowserImpl::getlang() throw (RuntimeException)
{
    return current_lang;
}


void ObjectBrowserImpl::MyWin::Resize()
{
    const unsigned int short padding = 5;
    const unsigned int short margin = 10;
    Size treeSize = WorkWindow::GetOutputSizePixel();
    treeSize.setHeight(treeSize.getHeight() - lang_uno.CalcMinimumSize().getHeight()*2 - 3*padding - 2*margin);
    treeSize.setWidth(treeSize.getWidth() - 2*margin);
    tree.SetPosPixel(Point(margin, margin));
    tree.SetSizePixel(treeSize);
    Point unoPos(margin, treeSize.getHeight() + 2*padding);
    lang_uno.SetPosPixel(unoPos);
    lang_uno.SetSizePixel(lang_uno.CalcMinimumSize());
    Point basPos(margin, unoPos.Y() + lang_uno.GetSizePixel().getHeight() + padding);
    lang_basic.SetPosPixel(basPos);
    lang_basic.SetSizePixel(lang_basic.CalcMinimumSize());
}


ObjectBrowserImpl::ObjectBrowserImpl(const Reference<XComponentContext> & xContext):
    m_xContext(xContext),
    w(this), 
    tree(w.tree),
    reflection(getCoreReflectionImpl(*m_xContext.get())),
    introspection(getIntrospectionImpl(m_xContext)),
    current_lang(ObjectBrowserLang_UNO)
{ }


void SAL_CALL ObjectBrowserImpl::insertByName(const OUString& name, const Any& obj)
    throw (IllegalArgumentException, ElementExistException, 
	   WrappedTargetException, 
	   RuntimeException)
{
    using std::pair;
    Guard<Mutex> guard(mutex);

    Reference<XTypeProvider> obj_xt;
    {
	obj >>= obj_xt;
	if (!obj_xt.is()) {
	    throw IllegalArgumentException();
	}
    }
    typedef std::hash_map<OUString, object_data, OUStringHash>::iterator iterator;
    pair<iterator,bool> insert_result = objects.insert(std::make_pair(name, object_data() ));
    if (!insert_result.second) {
	throw ElementExistException();
    }
    struct object_data& od = insert_result.first->second;
    od.obj = &obj;
    od.tree_item = tree.InsertEntry(XubString(name));
    od.inspect = introspection->inspect(obj);
    {
	Sequence<Type> interfaces = obj_xt->getTypes();
	od.interfaces = Sequence<Reference<XIdlClass> > (interfaces.getLength() + 1);
	od.interfaces[0] = TypeToIdlClass(XInterface::static_type(), *m_xContext.get());
	for (unsigned int i = 0; i < (unsigned int) interfaces.getLength(); i++) {
	    const Type& interface = interfaces[i];
	    const Reference<XIdlClass> interface_refl = TypeToIdlClass(interface, *m_xContext.get());
	    od.interfaces[i + 1] = interface_refl;
	}
    }
    {
	Reference<XServiceInfo> obj_srv_info;
	*od.obj >>= obj_srv_info;
	if (obj_srv_info.is()) {
	    od.services = obj_srv_info->getSupportedServiceNames();
	}
    }

    createMethodMap(od);
    fillTree(od);
}

void ObjectBrowserImpl::fillTree(ObjectBrowserImpl::object_data& od)
{
    SvLBoxEntry* methods_branch = tree.InsertEntry(XubString(RTL_CONSTASCII_USTRINGPARAM("methods")), od.tree_item);
    {
	Sequence<Reference<XIdlMethod> > m = od.inspect->getMethods(com::sun::star::beans::MethodConcept::ALL);
	for (unsigned int i = 0; i < (unsigned int) m.getLength(); i++) {
	    Reference<XIdlClass> interface;
	    ObjectBrowserImpl::method_map_t::iterator iface_it = od.method_map.find(m[i]->getName());
	    if (iface_it != od.method_map.end()) {
		interface = iface_it->second;
	    }
	    SvLBoxEntry* tree_entry = langs[current_lang]->insertMethod(tree, methods_branch, m[i]);
	    if (interface.is()) {
		tree_entry->AddItem(new FieldData(interface->getName(), m[i]->getName(), &od, TRUE));
	    }
	}
    }
    
    SvLBoxEntry* properties_branch = tree.InsertEntry(XubString(RTL_CONSTASCII_USTRINGPARAM("properties")), od.tree_item);
    {
	Sequence<Property> p = od.inspect->getProperties(com::sun::star::beans::MethodConcept::ALL);
	for (unsigned int i = 0; i < (unsigned int) p.getLength(); i++) {
	    SvLBoxEntry* tree_entry = langs[current_lang]->insertProperty(tree, properties_branch, p[i]);
	}
    }


    {
	if (od.services.getLength() > 0) {
	    SvLBoxEntry* service_branch = tree.InsertEntry(XubString(RTL_CONSTASCII_USTRINGPARAM("services")), od.tree_item);
	    for (unsigned int i = 0; i < (unsigned int) od.services.getLength(); i++) {
		SvLBoxEntry* tree_entry = tree.InsertEntry(XubString(od.services[i]), service_branch);
		tree_entry->AddItem(new FieldData(od.services[i], OUString(), &od, FALSE));
	    }
	}
	
    }

    SvLBoxEntry* interfaces_branch = tree.InsertEntry(XubString(RTL_CONSTASCII_USTRINGPARAM("interfaces")), od.tree_item);
    od.interface_to_entry.clear();
    for (unsigned int i = 0; i < (unsigned int) od.interfaces.getLength(); i++) {
	insertInterface(tree, interfaces_branch, od.interfaces[i], od, current_lang);
    }
}


static bool is_method_in_parent(const OUString& name, const Reference<XIdlClass>& c, bool toplevel)
{
    if (!toplevel) {
	Reference<XIdlMethod> m = c->getMethod(name);
	if (m.is()) {
	    return true;
	}
    }
    Sequence<Reference<XIdlClass> > children = c->getSuperclasses();
    for (unsigned int i = 0; i < children.getLength(); i++) {
	bool in_this_child = is_method_in_parent(name, children[i], false);
	if (in_this_child) {
	    return true;
	}
    }
    return false;
}

void createMethodMap(struct ObjectBrowserImpl::object_data& od)
{
    for (unsigned int i = 0 ; i < od.interfaces.getLength(); i++) {
	Reference<XIdlClass>& r = od.interfaces[i];
	Sequence<Reference<XIdlMethod> > m = r->getMethods();
	for (unsigned int j = 0; j < m.getLength(); j++) {
	    ObjectBrowserImpl::method_map_t::iterator old_it = od.method_map.find(m[j]->getName());
	    bool do_insert;
	    if (old_it != od.method_map.end()) {
		// Check that the method is not defined in any parent
		do_insert = !is_method_in_parent(m[j]->getName(), r, true);
	    } else {
		do_insert = true;
	    }
	    if (do_insert) {
		if (old_it != od.method_map.end()) {
		    od.method_map.erase(old_it);
		}
		od.method_map.insert(std::make_pair(m[j]->getName(), r));
	    }
	}
    }
}



void SAL_CALL ObjectBrowserImpl::removeByName(const OUString& name)
    throw (NoSuchElementException, WrappedTargetException, 
	   RuntimeException)
{
    typedef std::hash_map<OUString, object_data, OUStringHash>::iterator hash_iterator;
    Guard<Mutex> guard(mutex);
    hash_iterator hash_iter = objects.find(name);
    struct object_data& od = hash_iter->second;
    tree.GetModel()->Remove(od.tree_item);
    objects.erase(hash_iter);
}

void SAL_CALL ObjectBrowserImpl::replaceByName(const OUString& name, const Any& element)
    throw (IllegalArgumentException, NoSuchElementException, 
	   WrappedTargetException, RuntimeException)
{
    removeByName(name);
    insertByName(name, element);
}

Any SAL_CALL ObjectBrowserImpl::getByName(const OUString& aName) 
    throw (NoSuchElementException, WrappedTargetException, 
	   RuntimeException)
{
    typedef std::hash_map<OUString, object_data, OUStringHash>::iterator hash_iterator;
    Guard<Mutex> guard(mutex);
    hash_iterator hash_iter = objects.find(aName);
    if (hash_iter == objects.end()) {
	throw NoSuchElementException();
    }
    struct object_data& od = hash_iter->second;
    return *od.obj;
}

Sequence<OUString> SAL_CALL ObjectBrowserImpl::getElementNames() 
    throw (RuntimeException)
{
    Guard<Mutex> guard(mutex);
    Sequence<OUString> result(objects.size());
    { 
	typedef std::hash_map<OUString, object_data, OUStringHash>::iterator iterator;
	iterator ih;
	unsigned int i;
	for (i = 0, ih = objects.begin(); ih != objects.end(); i++, ih++) {
	    result[i] = ih->first;
	}
    }
    return result;
}

sal_Bool SAL_CALL ObjectBrowserImpl::hasByName(const OUString& aName) 
    throw(RuntimeException)
{
    typedef std::hash_map<OUString, object_data, OUStringHash>::iterator iterator;
    using std::pair;
    Guard<Mutex> guard(mutex);
    iterator find_result = objects.find(aName);
    return (sal_Bool) (find_result != objects.end());
}


Type SAL_CALL ObjectBrowserImpl::getElementType() throw (RuntimeException)
{
    return getCppuVoidType();
}

sal_Bool SAL_CALL ObjectBrowserImpl::hasElements() throw (RuntimeException)
{
    Guard<Mutex> guard(mutex);
    return objects.size() != 0;
}


void insertInterface(SvTreeListBox& t, SvLBoxEntry* parent, const Reference<XIdlClass>& c, struct ObjectBrowserImpl::object_data& od, ObjectBrowserLang current_lang)
{
    SvLBoxEntry* base = t.InsertEntry(XubString(c->getName()), parent);
    od.interface_to_entry.insert(std::make_pair(c->getName(), base));
    {
	Sequence<Reference<XIdlMethod> > m = c->getMethods();
	if (m.getLength() > 0) {
	    SvLBoxEntry* methods_branch = t.InsertEntry(XubString(RTL_CONSTASCII_USTRINGPARAM("methods")), base);
	    for (unsigned int i = 0; i < (unsigned int) m.getLength(); i++) {
		SvLBoxEntry* tree_entry = langs[current_lang]->insertMethod(t, methods_branch, m[i]);
		tree_entry->AddItem(new FieldData(c->getName(), m[i]->getName(), &od, FALSE));
	    }
	}
    }
    {
	Sequence<Reference<XIdlField> > a = c->getFields();
	if (a.getLength() > 0) {
	    SvLBoxEntry* attributes_branch = t.InsertEntry(XubString(RTL_CONSTASCII_USTRINGPARAM("attributes")), base);
	    for (unsigned int i = 0; i < (unsigned int) a.getLength(); i++) {
		SvLBoxEntry* tree_entry = langs[current_lang]->insertAttribute(t, attributes_branch, a[i]);
		tree_entry->AddItem(new FieldData(c->getName(), a[i]->getName(), &od, FALSE));
	    }
	}
    }
}


/* Code copied and adapcted from the basic interpreter basic/source/classes/sbunoobj.cxx  */

static Reference<XIdlReflection> getCoreReflectionImpl(XComponentContext& xContext)
{
	static Reference<XIdlReflection> xCoreReflection;

	// Haben wir schon CoreReflection, sonst besorgen
	if( !xCoreReflection.is() )
	{
		xContext.getValueByName(
		    OUString( RTL_CONSTASCII_USTRINGPARAM("/singletons/com.sun.star.reflection.theCoreReflection") ) )
			>>= xCoreReflection;
		OSL_ENSURE( xCoreReflection.is(), "### CoreReflection singleton not accessable!?" );
		if( !xCoreReflection.is() )
		{
			throw DeploymentException(
				OUString( RTL_CONSTASCII_USTRINGPARAM("/singletons/com.sun.star.reflection.theCoreReflection singleton not accessable") ),
				Reference< XInterface >() );
		}
	}
	return xCoreReflection;
}


static Reference<XIntrospection> getIntrospectionImpl(Reference<XComponentContext> xContext)
{
    return createInstance<XIntrospection>(xContext, "com.sun.star.beans.Introspection");
}


/* Idem. This should not be neccessary. Filled bug # 66635  */

// TODO: Spaeter auslagern
static Reference<XIdlClass> TypeToIdlClass(const Type& rType, XComponentContext& xContext)
{
	// void als Default-Klasse eintragen
	Reference<XIdlClass> xRetClass;
	typelib_TypeDescription * pTD = 0;
	rType.getDescription( &pTD );

	if( pTD )
	{
		OUString sOWName( pTD->pTypeName );
		Reference< XIdlReflection > xRefl = getCoreReflectionImpl(xContext);
		xRetClass = xRefl->forName( sOWName );
	}
	return xRetClass;
}

}
